/*****************************************************************************
Copyright 1987, 1988, 1989, 1990, 1991 by Digital Equipment Corp., Maynard, MA

Permission to use, copy, modify, and distribute this software and its 
documentation for any purpose and without fee is hereby granted, 
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in 
supporting documentation, and that the name of Digital not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.  

DIGITAL DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
DIGITAL BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.

*****************************************************************************/
#include "Xlib.h"
#define NEED_REPLIES
#define NEED_EVENTS
#include "Xproto.h"

#include "xtraplib.h"
#include "xtraplibproto.h"
#ifndef vms
#include "Xlibint.h"
#else   /* vms */
#define SyncHandle() \
	if (dpy->synchandler) (*dpy->synchandler)(dpy)
/*
 * LockDisplay uses an undocumented feature in V5 of VMS that allows
 * disabling ASTs without calling $SETAST.  A bit is set in P1 space
 * that disables a user mode AST from being delivered to this process.
 * 
 */
#define LockDisplay(dis) 			\
{   globalref char ctl$gb_soft_ast_disable;	\
    globalref char ctl$gb_lib_lock;		\
    globalref short ctl$gw_soft_ast_lock_depth;	\
    if ( ctl$gb_soft_ast_disable == 0 ) {	\
	ctl$gb_soft_ast_disable = 1;		\
	ctl$gb_lib_lock = 1;			\
	ctl$gw_soft_ast_lock_depth = 1;		\
    }						\
    else ctl$gw_soft_ast_lock_depth++;		\
}

/*
 * UnlockDisplay clears the AST disable bit, then checks to see if an
 * AST delivery attempt was made during the critical section.  If so,
 * reenable_ASTs is set, and $SETAST must be called to turn AST delivery
 * back on.
 * 
 * Note that it assumed that LockDisplay and UnlockDisplay appear in
 * matched sets within a single routine.
 */
#define UnlockDisplay(dis) 			\
{   globalref char ctl$gb_reenable_asts;	\
    globalref char ctl$gb_soft_ast_disable;	\
    globalref char ctl$gb_lib_lock;		\
    globalref short ctl$gw_soft_ast_lock_depth;	\
    if (!--ctl$gw_soft_ast_lock_depth) 		\
	if ( ctl$gb_lib_lock ) {		\
	    ctl$gb_lib_lock = 0;		\
    	    ctl$gb_soft_ast_disable = 0;	\
	    if (ctl$gb_reenable_asts != 0)	\
		sys$setast(1);			\
        }					\
}

#define WORD64ALIGN
#if defined(__STDC__) && !defined(UNIXCPP)
#define GetReq(name, req) \
        WORD64ALIGN\
        if ((dpy->bufptr + SIZEOF(x##name##Req)) > dpy->bufmax)\
                _XFlush(dpy);\
        req = (x##name##Req *)(dpy->last_req = dpy->bufptr);\
        req->reqType = X_##name;\
        req->length = (SIZEOF(x##name##Req))>>2;\
        dpy->bufptr += SIZEOF(x##name##Req);\
        dpy->request++

#else  /* non-ANSI C uses empty comment instead of "##" for token concatenation
*/
#define GetReq(name, req) \
        WORD64ALIGN\
        if ((dpy->bufptr + SIZEOF(x/**/name/**/Req)) > dpy->bufmax)\
                _XFlush(dpy);\
        req = (x/**/name/**/Req *)(dpy->last_req = dpy->bufptr);\
        req->reqType = X_/**/name;\
        req->length = (SIZEOF(x/**/name/**/Req))>>2;\
        dpy->bufptr += SIZEOF(x/**/name/**/Req);\
        dpy->request++
#endif
#endif /* vms */

/* should not be called directly by clients */
#ifdef FUNCTION_PROTOS
static int XEConfigRequest(XETC *tc)
#else
static int XEConfigRequest(tc)
    XETC *tc;
#endif
{
    int status = True;
    Display *dpy = tc->dpy;
    CARD32 X_XTrapConfig = tc->extOpcode;
    xXTrapConfigReq *reqptr;
    GetReq(XTrapConfig,reqptr);
    reqptr->minor_opcode = XETrap_Config;
    memcpy((char *)&(reqptr->config), (char *)&(tc->values.v), 
        sizeof(XETrapCfg));
    XFlush(dpy);
    SyncHandle();
    if (tc->dirty & TCStream)
    {   /* We need to complete the connection */
        status = s_accept(XETrapGetTCStream(tc));
    }
    tc->dirty = 0L; /* Configuration is no longer dirty */
    return(status);
}

/* Flush out any pending configuration */
#ifdef FUNCTION_PROTOS
int XEFlushConfig(XETC *tc)
#else
int XEFlushConfig(tc)
    XETC *tc;
#endif
{
    return((tc->dirty) ? XEConfigRequest(tc) : True);
}
#ifdef FUNCTION_PROTOS
int XEResetRequest(XETC *tc)
#else
int XEResetRequest(tc)
    XETC *tc;
#endif
{
    int status = True;
    Display *dpy = tc->dpy;
    CARD32 X_XTrap = tc->extOpcode;
    xXTrapReq *reqptr;
    status = XEFlushConfig(tc); /* Flushout any pending configuration first */
    if (status == True)
    {
        GetReq(XTrap,reqptr);
        reqptr->minor_opcode = XETrap_Reset;
        XFlush(dpy);
        SyncHandle();
    }
    return(status);
}

#ifdef FUNCTION_PROTOS
int XEGetVersionRequest(XETC *tc, XETrapGetVersRep *ret) 
#else
int XEGetVersionRequest(tc,ret) 
    XETC *tc;
    XETrapGetVersRep *ret;
#endif
{ 
    int status = True;
    Display *dpy = tc->dpy;
    CARD32 X_XTrap = tc->extOpcode;
    xXTrapReq *reqptr;
    xXTrapGetVersReply rep; 
    int numlongs = (sizeof(rep)-sizeof(xReply)+sizeof(long)-1)/sizeof(long);
    LockDisplay(dpy); 
    GetReq(XTrap,reqptr);
    reqptr->minor_opcode = XETrap_GetVersion;
    (void)_XReply(dpy,(xReply *)&rep,numlongs,xTrue); 
    SyncHandle(); 
    UnlockDisplay(dpy); 
    memcpy((char *)ret,&(rep.data),sizeof(XETrapGetVersRep)); 
    return(status);
}

#ifdef FUNCTION_PROTOS
int XEGetLastInpTimeRequest(XETC *tc, XETrapGetLastInpTimeRep *ret) 
#else
int XEGetLastInpTimeRequest(tc,ret) 
    XETC *tc;
    XETrapGetLastInpTimeRep *ret;
#endif
{ 
    int status = True;
    Display *dpy = tc->dpy;
    CARD32 X_XTrap = tc->extOpcode;
    xXTrapReq *reqptr;
    xXTrapGetLITimReply rep; 
    int numlongs = (sizeof(rep)-sizeof(xReply)+sizeof(long)-1)/sizeof(long);
    LockDisplay(dpy); 
    GetReq(XTrap,reqptr);
    reqptr->minor_opcode = XETrap_GetLastInpTime;
    (void)_XReply(dpy,(xReply *)&rep,numlongs,xTrue); 
    SyncHandle(); 
    UnlockDisplay(dpy); 
    memcpy((char *)ret,&(rep.data),sizeof(XETrapGetLastInpTimeRep)); 
    return(status);
}

#ifdef FUNCTION_PROTOS
int XEGetAvailableRequest(XETC *tc, XETrapGetAvailRep *ret) 
#else
int XEGetAvailableRequest(tc,ret) 
    XETC *tc;
    XETrapGetAvailRep *ret;
#endif
{ 
    int status = True;
    Display *dpy = tc->dpy;
    CARD32 X_XTrap = tc->extOpcode;
    xXTrapReq *reqptr;
    xXTrapGetAvailReply rep; 
    int numlongs = (sizeof(rep)-sizeof(xReply)+sizeof(long)-1)/sizeof(long);
    LockDisplay(dpy); 
    GetReq(XTrap,reqptr);
    reqptr->minor_opcode = XETrap_GetAvailable;
    (void)_XReply(dpy,(xReply *)&rep,numlongs,xTrue); 
    SyncHandle(); 
    UnlockDisplay(dpy); 
    memcpy((char *)ret,&(rep.data),sizeof(XETrapGetAvailRep)); 
    return(status);
}

#ifdef FUNCTION_PROTOS
int XEStartTrapRequest(XETC *tc)
#else
int XEStartTrapRequest(tc)
    XETC *tc;
#endif
{ 
    int status = True;
    Display *dpy = tc->dpy;
    CARD32 X_XTrap = tc->extOpcode;
    xXTrapReq *reqptr;
    status = XEFlushConfig(tc); /* Flushout any pending configuration first */

    /* Add our event handler for the XLib transport */
    (void)XETrapSetEventHandler(tc, XETrapData, XETrapDispatchXLib);

    if (status == True)
    {
        GetReq(XTrap,reqptr);
        reqptr->minor_opcode = XETrap_StartTrap;
        XFlush(dpy);
        SyncHandle(); 
        BitTrue(tc->values.tc_flags, XETCTrapActive);
    }
    return(status);
}
#ifdef FUNCTION_PROTOS
int XEStopTrapRequest(XETC *tc) 
#else
int XEStopTrapRequest(tc) 
    XETC *tc;
#endif
{ 
    int status = True;
    Display *dpy = tc->dpy;
    CARD32 X_XTrap = tc->extOpcode;
    xXTrapReq *reqptr;
    status = XEFlushConfig(tc); /* Flushout any pending configuration first */
    if (status == True)
    {
        GetReq(XTrap,reqptr);
        reqptr->minor_opcode = XETrap_StopTrap;
        XFlush(dpy);
        SyncHandle(); 
        BitFalse(tc->values.tc_flags, XETCTrapActive);
    }

    /* Remove our event handler for the XLib transport */
    (void)XETrapSetEventHandler(tc, XETrapData, NULL);

    return(status);
}

#ifndef _XINPUT
#ifdef FUNCTION_PROTOS
int XESimulateXEventRequest(XETC *tc, CARD8 type, CARD8 detail,
    CARD16 x, CARD16 y, CARD8 screen) 
#else
int XESimulateXEventRequest(tc,type,detail,x,y,screen) 
    XETC  *tc;
    CARD8 type;
    CARD8 detail;
    CARD16 x,y;
    CARD8 screen;
#endif
{
    int status = True;
    Display *dpy = tc->dpy;
    CARD32 X_XTrapInput = tc->extOpcode;
    xXTrapInputReq *reqptr;
    status = XEFlushConfig(tc); /* Flushout any pending configuration first */
    if (status == True)
    {   /* Is this stream I/O (e.g. non-XLib input)? */
        if ((XETrapGetTCStream(tc)) && 
            (XETrapGetTCReadTransport(tc) != FamilyXLib))
        {   /* yes, so use the stream write */
            XETrapInputReq input;
            input.type     = type;
            input.detail   = detail;
            input.x        = x;
            input.y        = y;
            input.screen   = screen;
            if (s_write(XETrapGetTCStream(tc), (char *)&input, 
                sizeof(XETrapInputReq)) != sizeof(XETrapInputReq))
            {
                status = False;
            }
        }
        else
        {   /* no, use XLib transport for the write */
            GetReq(XTrapInput,reqptr);
            reqptr->minor_opcode = XETrap_SimulateXEvent;
            reqptr->input.type   = type;
            reqptr->input.detail = detail;
            reqptr->input.x      = x;
            reqptr->input.y      = y;
            reqptr->input.screen = screen;
            XFlush(dpy);
        }
    }
    return(status);
}
#endif
#ifdef FUNCTION_PROTOS
int XEGetCurrentRequest(XETC *tc, XETrapGetCurRep *ret) 
#else
int XEGetCurrentRequest(tc, ret) 
    XETC *tc;
    XETrapGetCurRep *ret;
#endif
{ 
    int status = True;
    Display *dpy = tc->dpy;
    CARD32 X_XTrap = tc->extOpcode;
    xXTrapReq *reqptr;
    xXTrapGetCurReply rep; 
    int numlongs = (sizeof(rep)-sizeof(xReply)+sizeof(long)-1)/sizeof(long);
    status = XEFlushConfig(tc); /* Flushout any pending configuration first */
    if (status == True)
    {
        LockDisplay(dpy); 
        GetReq(XTrap,reqptr);
        reqptr->minor_opcode = XETrap_GetCurrent;
        (void)_XReply(dpy,(xReply *)&rep,numlongs,xTrue); 
        SyncHandle();
        UnlockDisplay(dpy); 
        memcpy((char *)ret,&(rep.data),sizeof(XETrapGetCurRep)); 
    }
    return(status);
}

#ifdef FUNCTION_PROTOS
int XEGetStatisticsRequest(XETC *tc, XETrapGetStatsRep *ret) 
#else
int XEGetStatisticsRequest(tc,ret) 
    XETC *tc;
    XETrapGetStatsRep *ret;
#endif
{ 
    int status = True;
    Display *dpy = tc->dpy;
    CARD32 X_XTrap = tc->extOpcode;
    xXTrapReq *reqptr;
    xXTrapGetStatsReply rep; 
    int numlongs = (sizeof(rep)-sizeof(xReply)+sizeof(long)-1)/sizeof(long);
    status = XEFlushConfig(tc); /* Flushout any pending configuration first */
    if (status == True)
    {
        LockDisplay(dpy); 
        GetReq(XTrap,reqptr);
        reqptr->minor_opcode = XETrap_GetStatistics;
        (void)_XReply(dpy,(xReply *)&rep,numlongs,xTrue); 
        SyncHandle(); 
        UnlockDisplay(dpy); 
        memcpy(ret,&(rep.data),sizeof(XETrapGetStatsRep)); 
    }
    return(status);
}
