/*****************************************************************************
Copyright 1987, 1988, 1989, 1990, 1991 by Digital Equipment Corp., Maynard, MA

Permission to use, copy, modify, and distribute this software and its 
documentation for any purpose and without fee is hereby granted, 
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in 
supporting documentation, and that the name of Digital not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.  

DIGITAL DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
DIGITAL BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.

*****************************************************************************/
/*
 *
 *  CONTRIBUTORS:
 *
 *      Dick Annicchiarico
 *      Robert Chesler
 *      Dan Coutu
 *      Gene Durso
 *      Marc Evans
 *      Alan Jamison
 *      Mark Henry
 *      Ken Miller
 *
 */
#include "stdio.h"
#include "xtraplib.h"
#include "xtraplibproto.h"
#if defined VMS && defined DNETCONN            /* to the rest of the file!!! */
#include "descrip.h"
#include "file.h"
#include "iodef.h"
#include "libdef.h"
#include "dvidef.h"
#include "psldef.h"
#include "jpidef.h"
#include "ssdef.h"
#include "stsdef.h"
#include "nfbdef.h"
#include "prvdef.h"     /* Process privilege defs and constants */
#include "errno.h"
#include "msgdef.h"

#define InitDescriptorR(dsc_name, dsc_size, dsc_buff)             \
    dsc_name.dsc$w_length   = dsc_size;                           \
    dsc_name.dsc$b_dtype    = DSC$K_DTYPE_T;                      \
    dsc_name.dsc$b_class    = DSC$K_CLASS_S;                      \
    dsc_name.dsc$a_pointer  = dsc_buff;

    /* Misc. structures for declaring network objects */
typedef struct _nfb 
{
    char b_func;
    long l_terminator;
} NFB;

#define MAX_NCB     110
#define LOGNAM_SIZE 20

/* Net Command Mailbox data struct */
typedef struct
{
        INT16 msg;
        INT16 unit;
        char  name_info[MAX_NCB];
} NETCMD;

typedef struct dsc$descriptor_s _align (QUADWORD) dsc_quad;

extern QRECPtr s_get_free_buffer();
static $DESCRIPTOR ( dsc_net_device, "_NET:" );

/*
 *  DESCRIPTION:
 *
 *      This function performs the VMS DECnet initialization.  This 
 *      involves assigning a channel, declaring a network object 
 *      XESTREAMDDDDddddd (where "DDDD" is the process index to guarantee 
 *      uniqueness on a given node and "ddddd" is the I/O channel to 
 *      guarantee uniqueness within a given process), declaring an
 *      event flag for asynchronous operations, and then finally
 *      posting an asynchronous network access request.
 *
 */
#ifdef FUNCTION_PROTOS
int OpenDECnetStream(XESTREAM *st, char *addr, CARD32 mode)
#else
int OpenDECnetStream(st, addr, mode)
    XESTREAM *st;
    char     *addr;
    CARD32   mode;
#endif
{
    int vms_status;
    long  proc_index = 0L;
    short ret_cnt    = 0L;
    char  object_str[13L], tmpstr1[9L], tmpstr2[9L];
    char  *tmp, *ptr;
    /* Network control block to declare an object name. */
    static dsc_quad *quadptr;
    static struct dsc$descriptor_s _align (QUADWORD) dsc_object_str;
    static struct dsc$descriptor_s _align (QUADWORD) dsc_nfb;
    NFB  nfb = { NFB$C_DECLNAME, NULL };

    /* Allocate & Initialize ncb descriptor.  Length is updated in netcmd AST 
     * & freed when connection is lost/aborted.
     */
    if ((quadptr = (dsc_quad *)malloc(sizeof(dsc_quad))) == NULL)
    {
        fprintf(stderr,"SIO:  Can't allocate memory in OpenDECnetStream()!\n");
        return(False);
    }
    if ((quadptr->dsc$a_pointer = (char *)malloc(MAX_NCB)) == NULL)
    {
        fprintf(stderr,"SIO:  Can't allocate memory in OpenDECnetStream()!\n");
        return(False);
    }
    quadptr->dsc$w_length   = MAX_NCB;
    quadptr->dsc$b_dtype    = DSC$K_DTYPE_T;
    quadptr->dsc$b_class    = DSC$K_CLASS_S;
    st->u.Q.ncb = quadptr;
    InitDescriptorR(dsc_object_str, (sizeof(object_str)-1), object_str);
    InitDescriptorR(dsc_nfb, sizeof(nfb), &nfb);
    st->u.Q.read_mode = st->u.Q.write_mode = IO$_NOP;
    switch(mode)
    {   /* Tear down the read/write mode for DECnet */
        case O_RDONLY:
            st->u.Q.read_mode = IO$_READVBLK;
            break;
        case O_RDWR:
            st->u.Q.read_mode = IO$_READVBLK;
            /* fall through to get write as well */
        case O_WRONLY:
            st->u.Q.write_mode = IO$_WRITEVBLK;
            break;
        default:
            return(False);
            break;
    }
    /* Allocate & Initialize netcmd mbx logical descriptor.
     * Free'd when connection is lost/aborted.
     */
    if ((quadptr = (dsc_quad *)malloc(sizeof(dsc_quad))) == NULL)
    {
        fprintf(stderr,"SIO:  Can't allocate memory in OpenDECnetStream()!\n");
        return(False);
    }
    if ((quadptr->dsc$a_pointer = (char *)malloc(LOGNAM_SIZE)) == NULL)
    {
        fprintf(stderr,"SIO:  Can't allocate memory in OpenDECnetStream()!\n");
        return(False);
    }
    sprintf(quadptr->dsc$a_pointer, "MBX_%x", quadptr);
    quadptr->dsc$w_length   = strlen(quadptr->dsc$a_pointer);
    quadptr->dsc$b_dtype    = DSC$K_DTYPE_T;
    quadptr->dsc$b_class    = DSC$K_CLASS_S;
    st->u.Q.netcmd = quadptr;
    /* In the case of DECnet, the *port* refers to the network connection
     * and the *read_chan* and *write_chan* refer to the same data channel.
     * mbx_chan is just a noop channel we need to satisfy NETCMD.
     */
    if ((vms_status = sys$crembx(0L, &st->u.Q.mbx_chan, 
        (sizeof(QREC) + st->u.Q.bufsiz), 0L, 0xFF0F, PSL$C_USER, 
        st->u.Q.netcmd)) != SS$_NORMAL)
    {
        fprintf(stderr,
            "SIO:  Err %%x%x creating net mailbox in OpenDECnetStream()!\n",
            vms_status);
        return(False);
    }
    if ((vms_status = sys$assign(&dsc_net_device,&st->u.Q.port,PSL$C_USER,
        st->u.Q.netcmd)) != SS$_NORMAL)
    {
        fprintf(stderr,
            "SIO:  Err %%x%x assigning channel in OpenDECnetStream()!\n",
            vms_status);
         return(False);
    }

    sprintf(tmpstr1, "%08x", getpid());
    sprintf(tmpstr2, "%03x", st->u.Q.mbx_chan);
    sprintf(object_str,"S%8.8s%3.3s", &(tmpstr1[strlen(tmpstr1)-8L]),
        &(tmpstr2[strlen(tmpstr2)-3L]));    /* get last digits */
    dsc_object_str.dsc$w_length = strlen(object_str);    
    /* Now declare the net object on the network channel */
    /* This requires SYSNAM priv's so elevate if possible */
/* Security section! Do NOT add any code unless you know what you're doing! */
    sys$setast(False);
    change_privs(True,1,PRV$M_SYSNAM);
    vms_status = sys$qiow(0L,st->u.Q.port,IO$_ACPCONTROL,st->u.Q.iosb,
        0L,0L,&dsc_nfb,&dsc_object_str,0L,0L,0L,0L);
    change_privs(False,1,PRV$M_SYSNAM);
    sys$setast(True);
/* End of Security section! */
    if (vms_status != SS$_NORMAL)
    {
        fprintf(stderr,
            "SIO:  Err %%x%x creating net object in OpenDECnetStream()!\n",
            vms_status);
        return(False);
    }
    if (st->u.Q.iosb[0] != SS$_NORMAL)
    {
        fprintf(stderr,
            "SIO:  Err %%x%x creating net object in OpenDECnetStream()!\n",
            st->u.Q.iosb[0]);
        return(False);
    }
    if (!(tmp = getenv("SYS$NODE")))
    {
        fprintf(stderr,
            "SIO:  Err %%x%x getting SYS$NODE in OpenDECnetStream()!\n",
            vms_status);
        return(False);
    }
    /* Allocate some memory for the destination string including the two
     * null-terminators.
     */
    st->addr2_lgth = strlen(tmp)+strlen(object_str)+2L;
    if ((st->addr2 = (char *)malloc(st->addr2_lgth))
        == NULL)
    {
        fprintf(stderr,"SIO:  Can't allocate memory in OpenDECnetStream()!\n");
        return(False);
    }
    strcpy(st->addr2, tmp);
    /*Now get rid of the "::" in the nodename */
    st->addr2[strcspn(st->addr2,":")] = '\0';  
    ptr = &st->addr2[strlen(st->addr2)]+1L;
    strcpy(ptr, object_str);    /* append null-terminated object str */

    /* No further work can be done beyond declaring the object
     * and posting the read to the netcmd mbx.  The "accept()" must be done 
     * *after* the XEConfigRequest is performed.
     */
    sys$clref(st->u.Q.accept_efn);
    return(post_netcmd_read(st));
}

#ifdef FUNCTION_PROTOS
static void ast_setefn(long efn)
#else
static void ast_setefn(efn)
    long efn;
#endif
{
    sys$setef(efn);
}

#ifdef FUNCTION_PROTOS
int AcceptDECnetStream(XESTREAM *st)
#else
int AcceptDECnetStream(st)
    XESTREAM *st;
#endif
{
    int vms_status;
    int status = True;
    char timer[8] = "0 ::30";      /* 30 second timer */
    static struct dsc$descriptor_s _align (QUADWORD) dsc_timer;
    static _align(quadword) unsigned long time_quad[2L];

    InitDescriptorR(dsc_timer, strlen(timer), timer);
    if ((vms_status = sys$assign(&dsc_net_device,&st->u.Q.write_chan,
        PSL$C_USER, st->u.Q.netcmd)) != SS$_NORMAL)
    {
        fprintf(stderr,
            "SIO:  Err %%x%x assigning channel in AcceptDECnetStream()!\n",
            vms_status);
        status = False;
    }

    /* Wait for efn to be set indicating that a connect request has
     * been delivered to the netcmd mbx.  Don't wait forever, though!
     * (set a wakeup timer first).
     */
    else if (((vms_status = sys$bintim(&dsc_timer,time_quad)) != SS$_NORMAL)
        || ((vms_status = sys$setimr(0L, time_quad, ast_setefn, 
        st->u.Q.accept_efn, 0L))!= SS$_NORMAL) || 
        ((vms_status = sys$waitfr(st->u.Q.accept_efn)) != SS$_NORMAL))
    {
        fprintf(stderr,
        "SIO:  Err %%x%x waiting for connection in AcceptDECnetStream()!\n",
            vms_status);
        status = False;
    }
    sys$cantim();   /* just in case it didn't fire */
    if (((vms_status = sys$qiow(0L,st->u.Q.write_chan,IO$_ACCESS,st->u.Q.iosb,
        0L,0L,0L,(dsc_quad *)st->u.Q.ncb,0L,0L,0L,0L)) != SS$_NORMAL) || 
        (st->u.Q.iosb[0] != SS$_NORMAL))
    {
        fprintf(stderr,
        "SIO:  Err %%x%x accessing net in AcceptDECnetStream()!\n",
            vms_status);
        fprintf(stderr,
        "      IOSB return value, %%x%x\n", st->u.Q.iosb[0]);
        status = False;
    }

    st->u.Q.read_chan = st->u.Q.write_chan; /* so that read_qio works */
    if (st->u.Q.read_mode == IO$_READVBLK)
    {
        if (queue_read_qio(st) != True)
        {   /* We're reading! */
            status = False;
        }
    }
    return(status);
}


/*
 *  DESCRIPTION:
 *
 *      This function performs the transport specific functions for 
 *      closing the DECnet connection.  This is different from CloseChan
 *      in that a QIO to de-access from the net is req'd.  Following
 *      the de-allocation, it just calls XECloseChan.
 *
 */
#ifdef FUNCTION_PROTOS
int CloseDECnetStream(XESTREAM *st)
#else
int CloseDECnetStream(st)
    XESTREAM *st;
#endif
{
    static dsc_quad *quadptr;
    int vms_status;
    int status = True;

    if ((vms_status = sys$qiow(0L,st->u.Q.write_chan,IO$_DEACCESS|IO$M_SYNCH,
        0L,0L,0L,0L,0L,0L,0L,0L,0L)) != SS$_NORMAL)
    {
        fprintf(stderr,
        "SIO:  Err %%x%x de-accessing net in CloseDECnetStream()!\n",
            vms_status);
        status = False;
    }
    if (!CloseQIOStream(st))       /* close the data channel */
    {
        status = False;
    }
     /* now close netcmd mbx chan */
    if ((vms_status = sys$dassgn(st->u.Q.port)) != SS$_NORMAL)
    {
        fprintf(stderr,
        "SIO:  Err %%x%x de-assigning net channel in CloseDECnetStream()!\n",
            vms_status);
        status = False;
    }
    /* now free ncb memory */
    quadptr = (dsc_quad *)st->u.Q.ncb;
    free(quadptr->dsc$a_pointer);
    free(quadptr);

    return(status);
}

#ifdef FUNCTION_PROTOS
static void ast_process_netcmd(QRECPtr qr)
#else
static void ast_process_netcmd(qr)
    QRECPtr qr;
#endif
{
    NETCMD   *nc      = (NETCMD *)qr->data;
    static dsc_quad *quadptr;
    CARD16 i;
    CARD16 start;
    CARD16 info_cnt;
    char *ncb;

    quadptr = (dsc_quad *)qr->st->u.Q.ncb;
    ncb = quadptr->dsc$a_pointer;
    switch(nc->msg)
    {
        case MSG$_ABORT:
        case MSG$_DISCON:
        case MSG$_EXIT:
        case MSG$_PATHLOST:
        case MSG$_PROTOCOL:
        case MSG$_THIRDPARTY:
        case MSG$_TIMEOUT:
        case MSG$_NETSHUT:
            CloseDECnetStream(qr->st);
            break;
        case MSG$_CONNECT:
            /* The following message connect munging is straight out
             * of old XTrap Code as well as in DB_SERVER.C DECnet sample.
             */
            i = nc->name_info[0] + 1L;
            info_cnt = nc->name_info[i];
            start = i + 1L;
            for (i=0; i<info_cnt; i++ )
            {
                ncb[i] = nc->name_info[start+i];
            }
            quadptr->dsc$w_length = info_cnt;

            sys$setef(qr->st->u.Q.accept_efn);
            break;
    }
    lib$insqti(qr, qr->st->u.Q.free_queue);
    if ((nc->msg == MSG$_CONNECT) && (qr->st->u.Q.port != -1L))
    {
        (void)post_netcmd_read(qr->st);  /* post another read */
    }

}

#ifdef FUNCTION_PROTOS
static int post_netcmd_read(XESTREAM *st)
#else
static int post_netcmd_read(st)
    XESTREAM *st;
#endif
{
    QRECPtr buf;
    int  status = True;
    int  vms_status;

    if ((!(buf = s_get_free_buffer(st))) || (st->u.Q.bufsiz < 110L))
    {
        /* Can't allocate read buffer! */
        fprintf(stderr, 
            "SIO:  Can't allocate memory in post_netcmd_read()!\n");
        st->u.Q.errno = ENOMEM;
        status = False;
    }
    else
    {
        buf->st = st;
        if ((vms_status = sys$qio(0L,st->u.Q.mbx_chan,IO$_READVBLK,buf->iosb,
            ast_process_netcmd,buf,buf->data,110L,0L,0L,0L,0L))
            != SS$_NORMAL)
        {
            fprintf(stderr,
                "SIO:  Error %%x%x returned in post_netcmd_read()!\n",
                vms_status);
            st->u.Q.errno = EIO;
            status = False;
        }
    }
    return(status);
}
#endif




