/*****************************************************************************
Copyright 1987, 1988, 1989, 1990, 1991 by Digital Equipment Corp., Maynard, MA

Permission to use, copy, modify, and distribute this software and its 
documentation for any purpose and without fee is hereby granted, 
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in 
supporting documentation, and that the name of Digital not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.  

DIGITAL DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
DIGITAL BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.

*****************************************************************************/
/*
 *
 *  CONTRIBUTORS:
 *
 *      Dick Annicchiarico
 *      Robert Chesler
 *      Dan Coutu
 *      Gene Durso
 *      Marc Evans
 *      Alan Jamison
 *      Mark Henry
 *      Ken Miller
 *
 */
#include "stdio.h"
#include "xtraplib.h"
#include "xtraplibproto.h"
#if defined VMS && defined MBXCONN            /* to the rest of the file!!! */
#include "descrip.h"
#include "file.h"
#include "iodef.h"
#include "libdef.h"
#include "dvidef.h"
#include "psldef.h"
#include "jpidef.h"
#include "ssdef.h"
#include "stsdef.h"
#include "nfbdef.h"

typedef struct
    {
        short buff_len;
        short code;
        long  *ret_info;
        long  ret_len;
        long  terminator;
    } GETDVI_ITM;

#ifdef FUNCTION_PROTOS
static int do_read_mbx(XESTREAM *st, char *mbx, GETDVI_ITM *getdvi_itm)
#else
static int do_read_mbx(st, mbx, getdvi_itm)
    XESTREAM *st;
    char     *mbx;
    GETDVI_ITM *getdvi_itm;
#endif
{
    int vms_status;
    st->u.Q.read_mode = IO$_READVBLK;
    /* Need to create a mailbox & then go ask what it's name is */
    if ((vms_status = sys$crembx(0L, &st->u.Q.read_chan, MbxSize, 
        MbxSize*MbxBufs, 0L, 0L, 0L)) != SS$_NORMAL)
    {
        fprintf(stderr,
            "SIO:  Err %%x%x creating mailbox in OpenMbxStream()!\n",
            vms_status);
        return(False);
    }
    if ((vms_status = sys$getdviw(0L,st->u.Q.read_chan,0L,
        getdvi_itm,0L,0L,0L,0L)) != SS$_NORMAL)
    {
        fprintf(stderr,
            "SIO:  Err %%x%x getting device name in OpenMbxStream()!\n",
            vms_status);
        return(False);
    }
    /* Allocate some memory for the destination string including the
     * null-terminator.
     */
    st->addr2_lgth = strlen(mbx)+2L; /* String and 2 '\0' */
    if ((st->addr2 = (char *)calloc(1L, st->addr2_lgth)) == NULL)
    {
        fprintf(stderr, "SIO:  Can't allocate memory in OpenMbxStream()!\n");
        return(False);
    }
    strcpy(st->addr2, mbx);
    return(True);
}
/*
 *  DESCRIPTION:
 *
 *      This function performs the VMS Mailbox initialization.
 *
 */
#ifdef FUNCTION_PROTOS
int OpenMbxStream(XESTREAM *st, char *mbx, CARD32 mode)
#else
int OpenMbxStream(st, mbx, mode)
    XESTREAM *st;
    char     *mbx;
    CARD32   mode;
#endif
{
    int vms_status;
    int status = True;
    int tmp_lgth;
    char mbx_name[64L];
    GETDVI_ITM getdvi_itm = {64L, DVI$_DEVNAM, mbx_name, 0L, 0L};

    st->u.Q.read_mode = st->u.Q.write_mode = IO$_NOP;
    st->addr2 = NULL;
    st->addr2_lgth = 0L;
    switch(mode)
    {   /* Tear down the read/write mode for Mailboxes */
        case O_RDONLY:
            status = do_read_mbx(st, mbx_name, &getdvi_itm);
            break;
        case O_RDWR:
            status = do_read_mbx(st, mbx_name, &getdvi_itm);
            /* fall through to get write as well */
        case O_WRONLY:
            if (status == True)
            {
                st->u.Q.write_mode = IO$_WRITEVBLK | IO$M_NOW;
                /* Need to create a mailbox & then go ask what it's name is */
                if ((vms_status = sys$crembx(0L, &st->u.Q.write_chan, MbxSize, 
                    MbxSize*MbxBufs, 0L, 0L, 0L)) != SS$_NORMAL)
                {
                    fprintf(stderr,
                    "SIO:  Err %%x%x creating mailbox in OpenMbxStream()!\n",
                        vms_status);
                    status = False;
                }
                else if ((vms_status = sys$getdviw(0L,st->u.Q.write_chan,0L,
                    &getdvi_itm,0L,0L,0L,0L)) != SS$_NORMAL)
                {
                    fprintf(stderr,
                "SIO:  Err %%x%x getting device name in OpenMbxStream()!\n",
                        vms_status);
                    status = False;
                }
                /* Allocate some memory for the destination string including
                 * null-terminator.
                 */
                if (status == True)
                {
                    tmp_lgth = st->addr2_lgth;
                    st->addr2_lgth += strlen(mbx_name)+1L;
                    if ((st->addr2 = (char *)realloc(st->addr2, 
                        st->addr2_lgth)) == NULL)
                    {
                        fprintf(stderr, 
                        "SIO:  Can't allocate memory in OpenMbxStream()!\n");
                        status = False;
                    }
                    else
                    {   
                        if (!tmp_lgth)
                        {   /* No output Mbx?, add a preceeding null term. */
                            st->addr2[0] = '\0';
                        }
                        strcpy(&st->addr2[strlen(st->addr2)+1L], mbx_name);
                    }
                }
            }
            break;
        default:
            status = False;
            break;
    }

    return(status);
}

#endif
